; The use of the stack in the next few routines is unusual, make sure you
; understand this technique fully before altering any of this code. This is the
; fastest way to undraw things.

; Basically, every routine which draws onto the vdu puts the information which
; will be required to undraw again onto a stack (cs:bp). The format of this is
; chosen so that we can later point the stack (ss:sp) at this and POP everything
; back into registers again, including using RET's to pop the undraw routine
; addresses.

; N.B. contary to popular opinion this can cope with interrupts, they simply add
; crap to the end of the new stack, this only overwrites data after it has been
; pulled off and used. (honest !) 

; This undraws everything in the undraw stack

UD_Col		equ 0000h		; Undraw to black

Undraw_Old:	push es			; Save
		push ds			;
 
		push di			;
		push si			;

		push dx			;
		push cx			;
		push bx			;
		push ax			;

		; Now point at the un-draw list

		cli			; DI, can't have an interrupt in here

		; Should we fuck around ?

		cmp b cs:Language,1	; C ?
		je UO_C_Hack		; Yes, skip

UO_Entry:	mov cs:UD_Stack_Seg,ss	; Save
		mov cs:UD_Stack_Ptr,sp	;

		mov ax,cs		; New stack
		mov ss,ax		;

		test b cs:DRAW_Page,-1	; Page 0 ?
		je UO_P0		; Yes, skip

		mov sp,cs:pUS_1		; Use the ptr to page 1

		jmp s UO_GotPtr		; Skip

	; Page 0

UO_P0:		mov sp,cs:pUS_0		; Use the ptr to page 0

	; Clear the page

UO_GotPtr:	sti			; EI

		; Now undraw everything

		ret			; This "calls" the last drawn item

	; Now, if we are emulating C, we should fuck-up occasionally !

UO_C_Hack:	call Rand_32		; Giz a random		

		and ax,3FFh
		cmp ax,2		; 2/1024 chance
		jae UO_Entry		; OK, undraw it

	; Just don't bother to undraw anything !

; This is the dummy sentinel routine

Undraw_End:	cli			; DI

		mov ss,cs:UD_Stack_Seg	; Restore the stack
		mov sp,cs:UD_Stack_Ptr	;

		sti			; EI

		; Now restore the regs

		pop ax			; Restore
		pop bx			;
		pop cx			;
		pop dx			;

		pop si			;
		pop di			;

		pop ds			;
		pop es			;

	; Now point BP at the current page's stack

		mov bp,o Undraw_Stack_0	; Assume page 0

		test b cs:DRAW_Page,-1	; Page 0 ?
		je UE_Done		; Yes, skip

		mov bp,o Undraw_Stack_1	; Page 1

	; Set this stack

UE_Done:	mov w cs:[bp],o Undraw_End ; Dummy sentinel routine

		ret			; Done

; Now setup the undraw stacks, and clear both pages

Init_Pages:	mov bp,o Undraw_Stack_0	; Point at the stack of items to undraw
		mov cs:pUS_0,bp		;

		mov w cs:[bp],o Undraw_End ; Dummy sentinel routine

		; And page 1

		mov bp,o Undraw_Stack_1	; Point at the stack of items to undraw
		mov cs:pUS_1,bp		;

		mov w cs:[bp],o Undraw_End ; Dummy sentinel routine

		; Set VDU Page 1, DRAW page 0

		mov b cs:DRAW_Page,1	; Set it

		; Done

		call SWAP_Pages		; Swap them

		call CLS_320		; Clear it

		call SWAP_Pages		; And the others

		jmp CLS_320		; Clear it

; When the drawing has ended, save the bp

Draw_End:	test b cs:DRAW_Page,-1	; Page 0 ?
		je DE_P0		; Yes, skip

		mov cs:pUS_1,bp		; Save the ptr to page 1

		ret			; Done

	; Page 0

DE_P0:		mov cs:pUS_0,bp		; Save ptr to page 0

		ret			; Done

; This code swaps the pages

Swap_Pages:	mov al,cs:DRAW_Page	; Get it

		call Set_DRAW_Page	; Drive it

		call Set_VDU_Page	; Show this one

		xor al,1		;
		mov cs:DRAW_Page,al	; Swap it

		ret			; Done

; Draw the players' ship, given di, and create un-draw entry on bp "stack"

Draw_Ship:	push ax			; Save
		push bx			;
		push dx			;

		; Calc X

		mov ax,cs:Scr_X		; LHS's X
		mov bx,cs:Ship_X	; Ship real X

		call Calc_Offset	; ax := difference

		mov cs:Map_X,ax		; Save for the radar

		mov dx,ax		; dx := x

		; Calc Y

		mov ah,cs:Ship_Y	; Y

		; Setup for the draw

		mov bx,0309h		; Colours
		mov cx,0B01h		;

		; Sprite number

DShip_0:	mov al,1		; Ship-sprite

		; Change sprite according to dirn

		test b cs:Ship_Dir,80h	; Left ?
		je DShip_1		; Yes, skip		

		; Sprite number

		mov al,2		; Ship-sprite

		; Draw it

DShip_1:	test b cs:Action,1	; Finished ?
		jnz DShip_3		; No, skip

		; Use the alternate sprite

		push ax			; Save
		push bx			;

		mov al,12		; Alternate sprite

		add ah,9		; Down

		; Flash the colours

		mov bl,cs:Frame_Cnt	; Get the count

		shr bl,1		;
		shr bl,1		;
		shr bl,1		;

		and bl,07h		; Mask 0 .. 7
		mov bh,bl		; Dup

		or bx,040Ch		;

DShip_2:	call Draw_Sprite	; Draw it

		sub ah,9		; Back

		pop bx			; Restore
		pop ax			;

		; Show the ship

DShip_3:	call Draw_Sprite	; Show it

		; Thrust required ?

		test b cs:Thrust_F,-1	; Under way ?
		jz DShip_5		; No, skip

		; Show the exhaust . . .

		mov al,42		; Ship-sprite

		; Change sprite according to dirn

		test b cs:Ship_Dir,80h	; Left ?
		je DShip_4		; Yes, skip		

		; Sprite number

		mov al,40		; Ship-sprite

		; Draw it

DShip_4:	push ax

		push dx
		call Rand_32
		pop dx

		mov bx,ax		; Dup
		and bx,0002h		; 0/2
		dec bx			; -1/+1
		add dx,bx		;

		mov bx,ax		; Dup
		and bx,00707h		;
		or bx,0040Ch		;

		mov cl,ah		; Dup
		and cl,01h		; 0/1

		pop ax

		add al,cl		; Change sprite		
		mov cx,060Ch		;

		call Draw_Sprite	; Show it

		; Now, show it in the radar

DShip_5:	mov ds,cs:VDU_Seg	; VDU Segment

		; Calc it's map posn 

		mov bl,cs:Ship_Y	; Y
		mov bh,0		; 16-bit

		add bx,bx		; * n2

		add bx,o Rad_Div_Tab - 2*Min_Y ; Look it up

		mov ax,cs:[bx]		; Look it up

		; Now the X

		mov bx,cs:Map_X		; X

		sar bx,1		; /8 (Scale to screen)
		sar bx,1		;
		sar bx,1		;

		add bx,160-20		; Offset it

		mov cl,bl		; Save
		and cl,07h		;

		sar bx,1		; / 8
		sar bx,1		;
		sar bx,1		;

		add bx,ax		; Allow for Y
		add bx,cs:VDU_Offset	; Offset

		; Calc bit mask

		mov ah,080h		; LHS
		shr ah,cl		; Align

		; Set the bit-mask

		mov dx,03CEh		; Set the bit-map

		mov al,08h		; BIT-MAP reg
		out dx,ax		; Set the map

		; Draw it

		and b [bx],0Fh		; Set the colour

		; Add to the un-draw list

		dec bp			; Push it
		dec bp			;

		mov cs:[bp],bx		; Save ptr

		; Now, exit

		pop dx			; Restore
		pop bx			;
		pop ax			;

		ret			; Done

; Draw an alien's ship, given di, and create un-draw entry on bp "stack"

DT_OffScrn:	pop bx			; Restore
		pop ax			;

		stc			; Failed to show it

		ret			; Done

	; Entry with di pointing at data

Draw_Thing:	push ax			; Save
		push bx			;

		mov ax,cs:Scr_X		; LHS's X
		mov bx,cs:d_X[di]	; Thing's real X

		call Calc_Offset	; ax := difference

		mov cs:Map_X,ax		; Save the Map X

		; Save it

		mov dx,ax		; dx := X

		; Is this valid ?

		add ax,31		; In range ?
		js DT_OffScrn		; No, skip

		cmp ax,320+31+31	; In range ?
		jnc DT_OffScrn		; No, skip

		; Calc Y

		mov ah,cs:d_Y[di]	; Y

		; Setup for the draw

		mov bx,cs:dCol1[di]	; Colours
		mov cx,cs:dCol2[di]	;

		; Sprite number

		mov al,cs:dShape[di]	; Thing

		; Change sprite according to dirn

		; Draw it

		call Draw_Sprite	; Draw it

		; Now, exit

		pop bx			; Restore
		pop ax			;

		clc			; Shown

		ret			; Done

; Draw the mountains, and put them into the undraw stack

Draw_Mount:	push ds			; Save
		push es			;

		push di			; Save
		push si			;

		push dx			;
		push cx			;
		push bx			;
		push ax			;

		; Set the address

		mov es,cs:VDU_Seg	; Segment

		add ax,ax		; * 2
		mov si,ax		;

		add si,o Mountains	; Start of mountains

		mov ax,DATA		; Setup DS
		mov ds,ax		;

		; Setup the required EGA modes, urgh !

		; Set Write mode 0, Read-mode 1

		mov dx,3CEh		; Graphics controller
		mov ax,00805h		; Mode/Reg
		out dx,ax		; Set it

		; Set the colour don't care register

		mov ax,00007h		; CDC := 0
		out dx,ax		;

		; Set the function to replace

		mov ax,00003h		; DRF := 0
		out dx,ax		;

		; Set the Enable Set/Reset to pass the Set/Reset bits

		mov ax,00F01h		; ESR := 0F
		out dx,ax		;

		; Now set the required colour

		mov al,00h		; Colour = 02 into Set/Reset
		mov ah,cs:Mount_Colour	;
		out dx,ax		;

		; Calc the vdu address

		push dx			; Save

		mov ax,189		; Calc Bottom-Left
		mov bx,cs:VDU_Width	;
		imul bx			;

		mov bx,cs:VDU_Offset	; Add LHS
		add bx,ax		;

		pop dx			; Restore

		; Point at the bottom-left

		mov ah,080h		; Pixel mask
		mov al,008h		; Mask register

		; Now, are we in the small or the large screen ?

		test b cs:Minature_F,-1	; Small or large
		jz DML_Lp1		; Large

		; Small window

DM_Lp1:		out dx,ax		; Set the bit-mask

		push ax			; Save
		push bx			;

		; Pass across for this bit of each byte

		mov cx,320/8		; This many

DM_Lp_I:	lodsw			; Get offset (ax)

		add si,14		; Skip

		add ax,bx		;
		mov di,ax		; Add base

		; Read and Write

		and es:[di],al		; Store it		
		and es:Y_Step_640[di],al; And below

		; Save this for the un-draw operation

		dec bp			; Move it
		dec bp			;

		mov cs:[bp],di		; Save ptr

		; Next

		inc bx			; Next byte

		loop DM_Lp_I		; Loop

		; Move ptr back

		sub si,(320/8)*16-2	; Point back

		; Now, do the next pixel posn.

		pop bx			; Mask
		pop ax			;

		ror ah,1		; Move mask		
		jnc DM_Lp1		; Loop

		; Setup the un-draw operation

DMS_X:		mov cs:-2[bp],es	; Ptr to screen for the un-draw

		mov w cs:-4[bp],o UD_Mount ; Routine for the undraw

		add bp,-4		; Move ptr

		; Done

		pop ax			; Restore
		pop bx			;
		pop cx			;
		pop dx			;

		pop si			;
		pop di			;

		pop es			;
		pop ds			;

		ret			; Done

	; Large window

DML_Lp1:	out dx,ax		; Set the bit-mask

		push ax			; Save
		push bx			;

		; Pass across for this bit of each byte

		mov cx,320/8		; This many

DML_Lp_I:	lodsw			; Get offset (ax)

		add si,14		; Skip

		add ax,bx		;
		mov di,ax		; Add base

		; Read and Write

		and es:[di],al		; Store it		
		and es:Y_Step_320[di],al; And below

		; Save this for the un-draw operation

		dec bp			; Move it
		dec bp			;

		mov cs:[bp],di		; Save ptr

		; Next

		inc bx			; Next byte

		loop DML_Lp_I		; Loop

		; Move ptr back

		sub si,(320/8)*16-2	; Point back

		; Now, do the next pixel posn.

		pop bx			; Mask
		pop ax			;

		ror ah,1		; Move mask		
		jnc DML_Lp1		; Loop

		; Exit

		jmp s DMS_X		; Exit

; Undraw the mountains, using the info. from the undraw stack.

UD_Mount:	pop ds			; Get the segment

		; Setup the required EGA modes, urgh !

		; Set Write mode 0, Read-mode 1

		mov dx,3CEh		; Graphics controller
		mov ax,00805h		; Mode/Reg
		out dx,ax		; Set it

		; Set the colour don't care register

		mov ax,00007h		; CDC := 0
		out dx,ax		;

		; Set the function to replace

		mov ax,00003h		; DRF := 0
		out dx,ax		;

		; Set the Enable Set/Reset to pass the Set/Reset bits

		mov ax,00F01h		; ESR := 0F
		out dx,ax		;

		; Now set the required colour

		mov ax,UD_Col		; Colour = 00 into Set/Reset
		out dx,ax		;

		; Now set the bit mask

		mov ax,0FF08h		; Mask = FF
		out dx,ax		;

		; Undraw them

		mov cx,320		; This many

		; Now, are we in the small or the large screen ?

		test b cs:Minature_F,-1	; Small or large
		jz UDML_Lp		; Large

		; Small

UDM_Lp:		pop bx			; Get the pixel address

		mov [bx],al		; Clear it, all 8 pixels
		mov Y_Step_640[bx],al	;

		loop UDM_Lp		; Loop

		; Done, undrawn !

		ret			; Goto the next undraw routine

	; Large

UDML_Lp:	pop bx			; Get the pixel address

		mov [bx],al		; Clear it, all 8 pixels
		mov Y_Step_320[bx],al	;

		loop UDML_Lp		; Loop

		; Done, undrawn !

		ret			; Goto the next undraw routine

; Vars

		align 2			; Even

VDU_Width	dw ?			; Offset to the next line

Map_X		dw ?			; The things X

UD_Stack_Seg	dw ?			; Save the stack segment
UD_Stack_Ptr	dw ?			; and offset

DRAW_Page	db ?			; Page we are drawing onto

Mount_Colour	db ?			; The mountain colour

		align 2			; Even

; Drawing stack for page 0

		dw 6000 dup (?)		; Space for storing undrawing data
Undraw_Stack_0	dw ?			; Top of stack

pUS_0		dw ?			; Ptr to stack

; Drawing stack for page 1

		dw 6000 dup (?)		; Space for storing undrawing data
Undraw_Stack_1	dw ?			; Top of stack

pUS_1		dw ?			; Ptr to stack

