o		equ offset		; Silly syntax
b		equ byte ptr		;
s		equ short		;

dos		EQU 21h			; DOS int
vdu		EQU 10h			; VDU int

Kbd_Status	EQU 64H			; Port address
Kbd_Data	EQU 60H			;

Int_Ctrl	EQU 20H			; Command
Int_Data	EQU 21H			; Data (Masks)

EOI		EQU 20H			; Command to 8259A

VDU_Status	EQU 3DAH		; Bit 0 = vertical sync

CODE		SEGMENT BYTE PUBLIC

		ASSUME  CS:CODE

; Given al,ah set the flags

Set_Frame_Flags:mov cs:Kill_FS,ah	; En/Dis frame lock
		mov cs:Int_Dis,al	; Int

		; Assert time-out if no frame-lock

		mov cs:Long_Time_F,ah	; Set if no frame sync

		ret			; Done

;  Wait for a frame

Next_Frame:	test b cs:Kill_FS,-1	; Sync ?
		jnz NF_X		; No, skip

		push dx			; Save regs
		push ax			;

		dec word ptr cs:Game_Time ; Time-out yet ?
		jnz NF_TO		; No, skip

		; Flag that we have been here a while

		mov al,1		; Set flag
		mov b cs:Long_Time_F,al	;

		; Wait a bit

NF_TO:		sti			; EI

		test b cs:Int_Dis,-1	; DI ?
		jnz NF_NI		; No, skip

		cli			; Disable ints

NF_NI: 		mov dx,VDU_Status	; Port address
		mov ah,08h		; The frame sync bit

NF_Lp:		in al,dx		; Look at VDU
		and al,ah		; Frame sync bit
		jnz NF_Lp		; Wait for end

NF_Lp1:		in al,dx		; Look at VDU
		and al,ah		; Frame sync bit
		jz NF_Lp1		; Wait for start

		sti			; Enable ints

		pop ax			; Restore
		pop dx			;

NF_X:		ret			; Done

Initiallise:	RET			; Nothing

SetUpDisplay:	MOV ah,0Fh		; Get video mode
		INT vdu			;

		MOV CS:OldMode,al	; Save mode
		MOV CS:OldPage,bh	; Save page

		MOV ax,0A000H		; EGA Page
		MOV CS:VDU_Seg,ax	; Save it
		MOV ds,ax		;

		mov al,3		; Colour

		mov ah,00h		; Text mode
		int vdu			;

;		; Clear border

;		mov ah,0Bh		;
;		mov bx,0		;

;		int vdu			;

		; Move cursor

;		mov ah,0fh		; Get page no.
;		int vdu			;

;		mov ah,2		;
;		mov dx,0		; LT

;		int vdu			;

		;

		call Set_640_350	; Front-end mode

		; Now, save the palette

		push cs			; Use the code-seg
		pop es			;

;		mov dx,o DAC_Save	; Save them here

		mov ax,01017h		; Read DAC values
		mov bx,00h		; All of them
		mov cx,100h		;

;		int vdu			; Read them

		; Now, set the new palette

;		mov dx,o Palette	; New values

		mov ax,01012h		; Write DAC values
		mov bx,00h		; All of them
		mov cx,100h		;

;		int vdu			; Read them

		; Done, exit

Wait_Settle:	MOV cx,0		; Pause

SUD_Wait:	PUSH ax			; Waste some time
		PUSH ax			;
		POP ax			;
		POP ax			;

		LOOP SUD_Wait		; Wait for display to settle

		; Return the VDU segment

		mov ax,cs:VDU_Seg	; Return it

		CLC			; OK
		RET			; Done

Kill_Vdu:	sti			; Enable ints

		MOV ah,0		; Restore video mode
		MOV al,CS:OldMode	;
		INT vdu

		MOV ah,5		; Restore page
		MOV al,CS:OldPage	;
		INT vdu			;

		JMP Wait_Settle		; Allow it to settle

CLS:		MOV bx,0		; Start of screen
		MOV cx,640/8*350	; Length of display
		XOR ax,ax		; Blank

CLS_Lp:		MOV [bx],al		; Clear it

		INC bx			;

		LOOP CLS_Lp		; Loop

		RET			; Done

CP_Fail:	MOV bx,0		; Dummy addr
		MOV ax,0FF00H		; Dummy mask and colour

		STC			; Fail
		RET			; Done

; Set the video modes

Set_640_350:	MOV al,010h		; EGA 640.350 16 colour

		MOV ah,0		; Set mode
		INT vdu			;

		ret			; Done

; Set the video modes

Set_320_200:	MOV al,00Dh		; EGA 320.200 16 colour

		MOV ah,0		; Set mode
		INT vdu			;

		ret			; Done

Rand_32:	PUSH cx			;
		PUSH bx			;

		CALL Rand_Inner		; Get next

		XOR al,dh		; Scramble it a bit
		XOR ah,dl		;

		POP bx			; Restore
		POP cx			;

		RET			; Done

Rand_Inner:	MOV ax,word ptr CS:Seed	; This is from Turbo Pascal
		MOV bx,word ptr CS:Seed+2;
		MOV cx,ax		;

		MUL word ptr CS:Nasty	;

		SHL cx,1		;
		SHL cx,1		;
		SHL cx,1		;

		ADD ch,cl		;
		ADD dx,cx		;
		ADD dx,bx		;

		SHL bx,1		;
		SHL bx,1		;

		ADD dx,bx		;
		ADD dh,bl		;

		MOV cl,5		;
		SHL bx,cl		;

		ADD dh,bl		;

		ADD ax,1		;
		ADC dx,0		;

		MOV word ptr CS:Seed,ax	;
		MOV word ptr CS:Seed+2,dx;

		RET			;

Nasty:		DW 33797		; Nasty number

Calc_Point:	CMP ch,200		; In range ?
		JAE CP_Fail		; No

		PUSH cx			; Save co-ords

		MOV dl,al		; Save colour

		XOR bx,bx		; Zero to begin with

CP_1:		MOV al,40		; Bytes per line
		MUL ch			; ax=Y offset

		ADD bx,ax		; Y done

		XOR ch,ch		; Lose silly bits

		ADD bx,cx		; X done

		POP cx			; Restore
		RET			; Done

; If called with al = NZ, then we use the AT keyboard

Link_Kbd:	mov dl,al		; Save the flag

		xor ax,ax		; Clear flags
		mov cs:Kbd_Flags,al	;

		; Clear the flags

		mov bx,o Key_Matrix	; Point at the keys
		mov cx,128		; This many

LK_Lp:		mov b cs:[bx],0		; Kill the flags
		inc bx			; Step the ptr

		loop LK_Lp		; Loop

		; Assume AT keyboard

		mov b cs:Keyb_Type,0	; Assert AT

		; Are we an AT ?

		mov ax,0F000h		; Seg = F000
		mov ds,ax		;

		mov al,ds:[0FFFEh]	; Get the byte from F000:FFFE

		cmp al,0FCh		; AT ?
		je LK_1			; Yes, skip

		; Now, are we forced to AT ?

		cmp dl,0		; Forced to AT ?
		jnz LK_1		; Yes, skip

		; Right, assume XT

		mov b cs:Keyb_Type,1	; Assert XT
		
		; Now get the vector, etc

LK_1:		push es			; Save segs
		push ds			;

		mov al,09h		; Kbd vector

		mov ah,035h		; Read vector
		int dos			;

		mov cs:Kbd_Old_Off,bx	; Save it
		mov bx,es		;
		mov cs:Kbd_Old_Seg,bx	; Save it

		push cs			; Get segment
		pop ds			;

		mov dx,o Kbd_Interrupt	; Assert it
		mov al,09h		; Keyboard int vector

		mov ah,25h		; Set interrupt vector
		int dos			;

		pop ds			; Restore
		pop es			;

		ret			; Done

Unlink_Kbd:	push ds			; Save ds

		mov dx,cs:Kbd_Old_Seg	; Get old vector
		mov ds,dx		;
		mov dx,cs:Kbd_Old_Off	;

		mov al,09h		; Kbd int

		mov ah,25h		; Restore it
		int dos			;

		pop ds			; Restore

		clc			; OK
		ret			; Done

; XT int

KI_XT:		push ax
		push bx

		in al,060h		; Read it
		mov bl,al		; Dup
		mov ah,al		; Dup

		in al,061h		; Ctrl

		or al,80h		; High
		out 61h,al		;
		and al,7Fh		; Low
		out 61h,al		;

		and bx,007Fh		; Point at the matrix
		add bx,o Key_Matrix	;

		rol ah,1		; Flag to bit 0
		not ah			; Invert state
		and ah,01h		; 0 = Not pressed, 1 = Pressed

		mov cs:[bx],ah		; Save key state

		; Exit

		mov al,61h		; Drive 8259A
		out 020h,al		;

		pop bx			; Restore
		pop ax			;

		iret			;

Kbd_Interrupt:	sti			; EI

		test b cs:Keyb_Type,1	; XT ?
		jnz KI_XT		; Yes, skip

		; AT type keyboard

		push ax			; Save regies
		push bx			;
		push cx			;
		push dx			;

		push di			;
		push si			;

		mov al,0ADh		; Disable keyboard
		call SendToKbd		; Ship it

		cli			; DI
		sub cx,cx		; Clear count

KI_Lp1:		in al,Kbd_Status	; Buffer full ?
		test al,02h		;
		loopnz KI_Lp1		; Loop

		in al,Kbd_Data		; Get the character from t'kbd
		sti			; Enable ints again

		cmp al,0FAh		; Is this an ACK ?
		je KI_Ack		; Yes, set the flag and ignore it.

		cmp al,0FEh		; Is this a resend ?
		je KI_Resend		; Yes, just ignore it (for now)

		mov bl,al		; Create ptr to Key Matrix
		and bx,007Fh		;
		add bx,o Key_Matrix	;

		rol al,1		; Flag to bit 0
		not al			; Invert state
		and al,01h		; 0 = Not pressed, 1 = Pressed

		mov cs:[bx],al		; Save key state

Kbd_I_Ret:	cli			; DI

		mov al,EOI		; End of interrupt
		out Int_Ctrl,al		; Tell 8259

		mov al,0AEh		; Enable kbd
		call SendToKbd		; Tell kbd

		cli			; DI

		pop si			; Restore regs
		pop di			;

		pop dx			;
		pop cx			;
		pop bx			;
		pop ax			;

		iret			; Done, restore flags

KI_Resend:	
KI_Ack:		or b cs:Kbd_Flags,01h	; Ack recieved
		jmp s Kbd_I_Ret		; Done

SendToKbd:	push ax			; Talk to Kbd

		cli			; DI
		sub cx,cx		; Clear count

STK_Lp:		in al,Kbd_Status	; Buffer full ?
		test al,02h		;
		loopnz STK_Lp		; Loop

		pop ax			; Restore data
		out Kbd_Status,al	; Send it

		sti			; EI
		ret			; Done

; Fixed data here

;		include Palette.inc	; Include the palette

;DAC_Save	db 256*3 dup (?)	; Saved DAC registers

Game_Time	dw 70*10		; About 10 seconds

VDU_Seg		DW 0			; Video segment
Kbd_Old_Off	DW 0			; Saved vector
Kbd_Old_Seg	DW 0			;

Kill_FS		db 0			; No sync
Int_Dis		db 0			; No DI's

Long_Time_F	db 0			; Not yet

Seed		DB "NIG!"		; Random number seed

Int_Mask	DB 0			; Initial int. mask

OldMode		DB 0			; Initial video mode
OldPage		DB 0			; Initial video page

Kbd_Flags	DB 0			; Some flags

Keyb_Type	db ?			; NZ = XT

Key_Matrix	DB 128 dup (0)		; Key matrix

CODE		ENDS

		PUBLIC SetUpDisplay
		PUBLIC Initiallise
		PUBLIC Link_Kbd
		PUBLIC Unlink_Kbd
		PUBLIC Kill_VDU
		PUBLIC Rand_32
		PUBLIC Calc_Point

		public Set_640_350
		public Set_320_200

		PUBLIC Next_Frame
		public Set_Frame_Flags

		PUBLIC Key_Matrix
		PUBLIC Seed

		public Long_Time_F

		END

