EOI		equ 20h			; Command to 8259A

; Drive the command line

Drive_Params:	mov si,0080H		; Get command line

		mov bl,ds:[si]		; Get the length
		mov bh,0		;

		inc si			; Point at the first

		add bx,si		; Point at the end

		mov b ds:[bx],0		; End the string
 
		; Find the first non space

		call SkipSpc		; Skip over length
		jz DPR_X		; EOL ? Yes , ignore

		; Look for a command

DPR_Lp:		call SkipSpc		; Avoid spaces

		mov di,o Commands	; Is it a command ?

		call FindCommand	; Find it
		jc DPR_Useage		; No, exit

		; We have a command

		mov b cs:HadOne,1	; Flag that we had a valid param.

		call bx			; call routine

		jmp s DPR_Lp		; Try another

	; Done, exit

DPR_X:		test b cs:fKill_Timer,-1; Kill the timer ?
		jz DPR_KT_X		; No, skip

		; Kill the 18.2 Hz timer

		push es			; Save segs
		push ds			;

		mov al,08h		; Timer vector

		mov ah,035h		; Read vector
		int dos			;

		mov cs:Timer_Old_Off,bx	; Save it
		mov bx,es		;
		mov cs:Timer_Old_Seg,bx	; Save it

		push cs			; Get segment
		pop ds			;

		mov dx,o Null_Int	; Assert it
		mov al,08h		; Timer int vector

		mov ah,25h		; Set interrupt vector
		int dos			;

		pop ds			; Restore
		pop es			;

DPR_KT_X:	ret			; Done

; If we grabbed it, set it again

Unlink_Timer:	test b cs:fKill_Timer,-1; Did we grab it ?
		jz UT_X			; No, skip

		push ds			; Save ds

		mov dx,cs:Timer_Old_Seg	; Get old vector
		mov ds,dx		;
		mov dx,cs:Timer_Old_Off	;

		mov al,08h		; Timer int

		mov ah,25h		; Restore it
		int dos			;

		pop ds			; Restore

UT_X:		ret			; Done

Null_Int:	push ax			; Save

		mov al,EOI		; End of interrupt
		out Int_Ctrl,al		; Tell 8259

		pop ax			; Restore

		iret			; Done, restore flags

	; Abort and show help

DPR_Useage:	test b cs:HadOne,1	; Had one ?
		jnz DPR_X		; Yes, play game

		mov dx,o sHelp		; Help string

	; Print a string and exit

PrintStrX:	call PrintStr		; Show it

		mov ah,terminate	; Back to dos
		int dos			;

; Look for a command

FindCommand:	mov dx,si		; Save pStart

FindLp:		cmp b cs:[di],0		; End of commands ?
		jz FindFail		; Yes, can't find one

		mov si,dx		; Restore pStart

FindA:		cmp b cs:[di],0		; End of command ?
		jz FindFound		; Yes

		mov al,ds:[si]		; Get next byte
		inc si			; Step ptr

		call CaseConv		; Make lc

		inc di			; Step ptr
		cmp al,cs:-1[di]	; Same char ?
		jz FindA		; Yes, loop

FindB:		inc di			; Find end of command $
		cmp b cs:-1[di],0	;
		jnz FindB		;

		add di,4		; Skip parameters

		jmp FindLp		; Try again

FindFail:	mov si,dx		; Restore ptr
		stc			; Assert fail
		ret			;

FindFound:	mov bx,cs:1[di]		; Read address of code
		mov dx,cs:3[di]		; Read ptr

		clc			; Assert ok

		ret			;

CaseConv:	cmp al,'A'		; Convert to lc
		jb CC_Done		;

		cmp al,'Z'		;
		ja CC_Done		;

		add al,'a'-'A'		; Convert

CC_Done:	ret			; Done		

; Print a string

PrintStr:	mov bx,dx		; Get ptr

PrintStr_Lp:	mov dl,cs:[bx]		; Get the char
		cmp dl,0		; End ?
		jz PSL_X		; Exit

		inc bx			; Skip the char

		mov ah,Display_Char	; Display it
		int dos			; Show it

		jmp s PrintStr_Lp	; Loop

PSL_X:		ret			; Done

; Set a flag

StFlag:		mov bx,dx		; Get offset

		mov b cs:[bx],1		; Assert flag

		ret			; Done

; Fixed data

Commands	db '?',0		;
		dw PrintStrX,sHelp	;

		db 'help',0		;
		dw PrintStrX,sHelp	;

		db '/int',0		; /int
		dw StFlag,fForce_Ints	;

		db '/timer',0		; /timer
		dw StFlag,fKill_Timer	;

		db '/ns',0		; /ns
		dw StFlag,fKill_FL	;

		db '/at',0		; /at
		dw StFlag,fForce_AT	;

		db '/quick',0		; /quick
		dw StFlag,fForce_Quick	;

		db '/high',0		; /high
		dw StFlag,fLoadHigh	;

		db '/keys',0		; /keys
		dw StFlag,fLoadKeys	;

		db '/opts',0		; /options
		dw StFlag,fLoadOptions	;

		db '/save',0		; /save
		dw StFlag,fWriteBack	;

		db 0			;

sHelp		db 'Commands include :-',0Dh
		db '/at	(Force AT,PS2 etc keyboard)',0Dh,0Ah
		db "	- you shouldn't need this, try it if your PC hangs !",0Dh,0Ah
		db '/int	(Alters the interrupt handler)',0Dh,0Ah
		db "	- try this to cure network timing problems, but the display will jerk !",0Dh,0Ah
		db '/timer	(Disable the 18.2 Hz interrupt)',0Dh,0Ah
		db "	- try this to cure display jerking problems",0Dh,0Ah
		db "/ns	(Don't frame lock)",0Dh,0Ah
		db "	- Try this as a joke, it goes faster !",0Dh,0Ah
		db "/quick	(Don't hide redraws in the front-end)",0Dh,0Ah
		db "	- Try this if the front-end screen changes are too slow for you",0Dh,0Ah
		db '/high	(Load highscores file)',0Dh,0Ah
		db '/opts	(Load options file)',0Dh,0Ah
		db '/keys	(Load keys file)',0Dh,0Ah
		db '/save	(Enables writing data on highscores, keys, etc on exit)',0Dh,0Ah
		db 0Ah
		db 'A sample command line to save the keys and restore them in future :-',0Dh,0Ah
		db 'ROTBS /save /high /keys /opts',0Dh,0Ah,0Ah
		db 'Have fun !'
		db 0dh,0ah,0

; Variables

fForce_AT	db 0		;

fLoadHigh	db 0		;
fLoadKeys	db 0		;
fLoadOptions	db 0		;
fWriteBack	db 0		;

fForce_Ints	db 0		;
fKill_FL	db 0		;
fKill_Timer	db 0		;

fForce_Quick	db 0		;

HadOne		db 0		; Not yet

Timer_Old_Seg	dw 0		; Timer vector
Timer_Old_Off	dw 0		; Timer vector
