; This is the sprite driving code. This creates a set of fast, optimised drawing
; code fragments, one for each sprite, and links them into a table.

nMax_Sprite	equ 100			; This many different shapes

cscMax_X	equ 64			; Size of the largest possible sprite
cscMax_Y	equ 30			;
cscY_Step	equ cscMax_X + 8	;

; Setup for un-drawing sprites

Set_Undraw_Sp:	dec bp			; Push marker
		dec bp			;

		mov w cs:[bp],0		; End of ship marker

		ret			; Done

; Save the undraw ptr

Finish_Undraw:	mov ax,cs:VDU_Seg	; The VDU segment
		mov cs:-2[bp],ax	; Save the segment

		sub bp,4		; Move down

		mov w cs:[bp],o UD_Sprite; Routine for the undraw

		ret			; Done

; Given al = sprite, dx = X, ah = Y, bx/cx = colours

Draw_Sprite:	push di			; Save
		push si			;

		push dx			; Save
		push ax			;

		; Save for later

		push ax			; Save

		xchg bx,si		; Save bx for later

		; Calc sprite ptr

		mov bl,al		; Get sprite no.
		dec bl			; Dec sprite number

		mov bh,0		; 16-bit

		add bx,bx		; * 8
		add bx,bx		;
		add bx,bx		;

		add bx,o Large_Sprites	; Point at the table
pSprite_Ptrs	equ this word - 2	; We modify this for speed

		; Now, allow for the offsets

		sub dx,cs:4[bx]		; Sub. X Offset
		sub ah,cs:6[bx]		; Sub. Y Offset

		cmp ah,200		; In range ?
		jae DS_NotVis		; No, this cannot be seen

		cmp ah,5		; In range ?
		jb DS_NotVis		; No, this cannot be seen

		; Now, calc the display ptr, clip value, offset etc

		push dx			; Save

		mov al,ah		; ax  := Y
		mov ah,0		; 0 .. 255

		mov dx,cs:VDU_Width	; Display YStep

		mul dx			; ax := Address fn(Y)

		mov di,ax		; Into di
		add di,cs:VDU_Offset	;

		pop dx			; Restore

		; Now, look at X, first the rotate value

		mov al,dl		; Bottom bits
		and al,07h		; Clip

		mov cs:DS_Rot,al	; Store it

		; Now the screen ptr

		sar dx,1		; / 8
		sar dx,1		;
		sar dx,1		;

		; dx = offset in bytes

		add di,dx		; Add to the screen ptr

		cmp dx,cs:VDU_Width	; Are we on the screen ?
		jb DS_On_Screen		; We are onscreen, skip

		jns DS_NotVis		; Not visible, skip

		; Ah. we are not on the screen, dx := Offset - 1

		dec dx			; dx := dx - 1

		jmp s DS_Clipped	; Skip

	; We are not visible at all, exit

DS_NotVis:	pop ax			;

		xchg bx,si		;

		pop ax			; Exit
		pop dx			;

		pop si			;
		pop di			;

		ret			; Done

	; We are onscreen, dx is 0 .. Y_Step-1 dx := 32767-dx

DS_On_Screen:	add dx,32767 - (320/8)	; dx := 32767 - Y_Step + Offset from LHS
	
	; Setup EGA/VGA registers

DS_Clipped:	push dx			; Save

		; Set Write mode 2, Read-mode 1

		mov dx,3CEh		; Graphics controller

		mov ax,00A05h		; Mode/Reg
		out dx,ax		; Set it

		; Set the colour don't care register

		mov ax,00007h		; CDC := 0F
		out dx,ax		;

		; Set the function to replace

		mov ax,00003h		; DRF := 0
		out dx,ax		;

		; Now calc the address, etc

		pop dx			; Restore
		pop ax			;

		; Stack the ptr to the routine

		xchg bx,si		; Swap ptr

		; Save the seg. regs

		push ds			; Save
		push es			;

		; Set the DS

		mov ds,cs:VDU_Seg	; The VDU segment

		; Pass this cs in es

		push cs			; es := cs
		pop es			;

		mov ah,cs:DS_Rot	; Rotate value

		call cs:dword ptr [si]	; Ptr at [si]

		; Recover various things

		pop es			; Restore
		pop ds			;

		pop ax			;
		pop dx			;

		pop si			;
		pop di			;

		ret			; Done

; This is the undraw routine for all the sprites

UD_Sprite:	pop ds			; Get the segment

		; Setup the required EGA modes, urgh !

		; Set Write mode 0, Read-mode 1

		mov dx,3CEh		; Graphics controller
		mov ax,00805h		; Mode/Reg
		out dx,ax		; Set it

		; Set the colour don't care register

		mov ax,00007h		; CDC := 0
		out dx,ax		;

		; Set the function to replace

		mov ax,00003h		; DRF := 0
		out dx,ax		;

		; Set the Enable Set/Reset to pass the Set/Reset bits

		mov ax,00F01h		; ESR := 0F
		out dx,ax		;

		; Now set the required colour

		mov ax,UD_Col		; Colour = 00 into Set/Reset
		out dx,ax		;

		; Now set the bit mask

		mov ax,0FF08h		; Mask = FF
		out dx,ax		;

		; Undraw the sprite

		mov ax,0		; 0000

		jmp s UDS_1		; Loop

	; Undraw one

UDS_Lp:		mov [bx],al		; Clear all 8 pixels

UDS_1:		pop bx			; Get the pixel address

		cmp bx,ax		; End ?
		jne UDS_Lp		; Yes, skip

		; Done, undrawn !

		ret			; Goto the next undraw routine

; Setup the pointer to the sprite set

Select_Sprites:	mov ax,o Large_Sprites	; Large
		mov bx,Y_Step_320	;

		test b cs:Minature_F,-1	; ?
		jz SS_Large		; Yes, skip

		mov ax,o Small_Sprites	; Small
		mov bx,Y_Step_640	;

SS_Large:	mov cs:pSprite_Ptrs,ax	; Save ptr
		mov cs:VDU_Width,bx	; Save the width


		ret			; Done

; Passed a ptr to the sprites in DI, create the drawing fns.

Gen_Sprites:	call Select_Sprites	; Select the ptr

	; Generate the sprites

GS_Lp:		mov ax,cs:[di]		; Get the sprite number

		cmp ax,0		; End ?
		je GS_Exit		; Yes, skip

		; Calc ptr to sprite ptr

		dec ax			; 0 .. n

		add ax,ax		; * 8
		add ax,ax		;
		add ax,ax		;

		add ax,cs:pSprite_Ptrs	; Base of table

		; Save ptr

		push ax			; Save

		; Now save offsets

		mov bl,cs:2[di]		; X offset
		mov cs:X_Off,bl		;

		mov bl,cs:3[di]		; Y offset
		mov cs:Y_Off,bl		;

		; Skip header

		add di,4		; Skip ID

		; Now, create this sprite in a buffer

		call Create_S_Code	; Generate the sprite at ds:0

		; Save the sprite ptr

		pop si			; Restore ptr to table

		mov w cs:[si],0		; Offset = 0
		mov cs:2[si],ds		; Segment = segment

		; Save the offsets

		mov al,cs:X_Off		; X offset
		cbw			; 16-bit
		mov cs:4[si],ax		; Save X offset

		mov al,cs:Y_Off		; Y offset
		mov cs:6[si],al		; Save Y offset

		; Sprite done, try again

		jmp s GS_Lp		; Try another one

	; Done, exit

GS_Exit:	ret			; Done

; Given di and the offsets, get some space and create a sprite

Create_S_Code:	call CSC_GetPixels	; Read the drawing into a buffer

		; Save the ptr

		push di			; Save

		; Now, take this data and generate the sprite

		mov si,0		; Offset = 0
		mov b cs:CSC_Flag,0	; Don't generate code

		; Try to generate code, to find the room required.

		call CSC_Inner		; Call it

		; Now, si = last byte, get some room.

		mov bx,si		; Size in bytes

		shr bx,1		; / 16
		shr bx,1		;
		shr bx,1		;
		shr bx,1		;

		inc bx			; + 16

		; Allocate this much

		mov ah,Allocate_Mem	; Call DOS
		int dos			;

		; Did it work ?

		jc CSC_NoRAM		; No, total failure . . .

		; Now, we must remember this block for later . . .

;		mov si,DATA		; Segment
;		mov ds,si		;

;		mov si,ds:pMemory_List	; Point at the list

;		mov ds:[si],ax		; Save the block
;		lea si,2[si]		; Skip

;		mov ds:pMemory_List,si	; Save ptr

;		mov w ds:[si],0		; Kill it

		; Now use this memory

		mov ds,ax		; Save the segment start

		; Now, create it again in this memory

		mov si,0		; Base = 0

		mov b cs:CSC_Flag,1	; Generate code

		call CSC_Inner		; Generate some code

		; Done, exit

CSC_NoRAM:	pop di			; Recover ptr

		ret			; Done, return ds

; Given di, etc. Generate the code to draw the sprite.

CSC_Inner:	mov di,o SC_Base	; Selection code

		mov cx,o SC_End-SC_Base	; Length

		; Now, copy this code

CSCI_Lp:	mov al,cs:[di]		; Get the code
		inc di			;

		call CSC_Poke		; Store it

		loop CSCI_Lp		; Loop

		; Now, build the drawing code

		mov di,o Inst_Table	; Point at some vars

		; Build one instance

CSCI_Lp2:	mov bx,cs:[di]		; Point at the pix buffer

		cmp bx,0		; End ?
		je CSCI_Finish		; Yes, skip

		; No, there is another instance of the sprite

		; Link this instance to the decode code.

		mov ax,si		; Get the current address

		push si			; Save ptr

		mov si,cs:2[di]		; Point at the link ptr

		; Build the jump instruction

		push ax			; Save

		mov al,0E9h		; JMP (d16)

		call CSC_Poke		; Store it

		pop ax			; Restore

		; Add the offset

		sub ax,si		; Calc offset
		sub ax,2		;

		call CSC_Poke_W		; [si],ax store the offset

		pop si			; Restore ptr

		; Create the code for this instance of the sprite

		push di			; Save

		mov di,bx		; Point at the pixel buffer

		call CSC_Instance	; Draw for this position

		pop di			; Restore

		; Next instance

		add di,4		; Skip to the next entry

		jmp s CSCI_Lp2		; Loop

	; We are done

CSCI_Finish:	clc			; Done
		ret			;

; Given di = ptr to Pix_Buff and si = PC we produce code.

CSC_Instance:	mov cx,cs:SC_Sizes	; cl = width, ch = height

		shr cl,1		; / 8
		shr cl,1		;
		shr cl,1		;

		add cl,2		; + 2 (CL = width in columns)

		; Create all columns

CSC_I_X_Lp:	push di			; Save ptr
		push dx			; Save offset
		push cx			; Save count

		; Save ptr to this head of column

		mov cs:CSC_TOC,si	; Save it

		mov b cs:CSC_Col_Code,0	; No code in this column yet

		mov b cs:CSC_Col_Mask,0	; No mask yet

		; Plant head of column

		mov bx,o HC_Base	; Selection code

		mov cx,o HC_End-HC_Base	; Length

		; Now, copy the head of column code

CSCI_X_Lp1:	mov al,cs:[bx]		; Get the code
		inc bx			;

		call CSC_Poke		; Store it

		loop CSCI_X_Lp1		; Loop

		; Restore column height

		pop cx			; Restore
		push cx			; Stack again

		; This is the point we have to plant an offset to the end.

		mov bp,si		; Save this point

		add si,2		; Skip the offset word

		; Setup the screen offset

		mov dx,0		; Offset to the top of the column

		mov cs:cOffset,dx	; Current offset

		; Now, generate this column as a set of lines

CSCI_Y_Lp:	push cx			; Save

		; Now, does this line need any drawing ?

		mov bx,di		; Point at this line

		mov cx,8		; This many pixels

		mov al,0		; None yet

CSCIY_Lp1:	or al,cs:[bx]		; Get next
		inc bx			;

		loop CSCIY_Lp1		; Loop

		; Any pixels ?

		cmp al,0		; Any
		je CSCIY_1		; No, skip

		; We have some pixels, calc the masks, etc

		call CSC_Line		; Generate one byte

		; Next line

CSCIY_1:	pop cx			; Restore

		; Move down the screen

		add dx,cs:VDU_Width	; Offset + screen offset per line

		; Move down the data

		add di,cscY_Step	; Down the data

		; Loop in Y

		dec ch			; One more line done
		jnz CSCI_Y_Lp		; Loop

		; This column is done, rewind di

		mov ax,cs:cOffset	; Offset from 0
		cmp ax,0		; Any ?
		jz CSCIY_2		; No, skip

		; Plant "SUB DI,nnn"

		push ax			; Save

		mov ax,0EF81h		; "SUB di,nnnn"
		call CSC_Poke_W		;

		pop ax			; Restore

		call CSC_Poke_W		; nnnn

		; We are done with this column, link to this point from above

CSCIY_2:	push si			; Save ptr

		mov ax,si		; Get it

		mov si,bp		; Restore ptr to link

		sub ax,si		; Calc displacement
		sub ax,2		; 

		; NOTE, we could check for offset = 0 here. (Why ?)

		call CSC_Poke_W		; Add the offset

		pop si			; Restore ptr

		; Now, is there any code in this column ?

		test b cs:CSC_Col_Code,1; Any code ?
		jne CSCIY_3		; Yes, skip

		; There is no code, don't have the TEST/JNS/JMP

		mov si,cs:CSC_TOC	; Rewind

		inc si			; Allow the INC SI

		; Now add the final "INC DI"

CSCIY_3:	mov al,047h		; "INC DI"
		call CSC_Poke		;

		; This column is done, next

		pop cx			; Restore
		pop dx			;
		pop di			;

		; Next column

		add di,8		; Across one column

		dec cl			; Loop
		jz CSCIY_4		; Fucking short branches

		jmp CSC_I_X_Lp		; Loop

		; Now, we are done. Replace the last "INC DI" with a "RETF"

CSCIY_4:	dec si			; Back one

		; End the column code

		push di			; Save

		mov di,o EX_Base	; Exit code

		mov cx,o EX_End-EX_Base	; Length

		; Now, copy this code

CSCIF_Lp:	mov al,cs:[di]		; Get the code
		inc di			;

		call CSC_Poke		; Store it

		loop CSCIF_Lp		; Loop

		pop di			; Recover

		; Exit

		ret			; Done

; Now generate one bytes worth of drawing from [di] to csc_poke.

		; Flag that we have had no EGA/VGA reads

CSC_Line:	mov b cs:CSC_Flag_VR,0	; No reads yet

		; Flag that we have had some code in this column

		mov b cs:CSC_Col_Code,1	; We have code

		; Now move the ptr

		mov ax,dx		; Required offset
		sub ax,cs:cOffset	; Difference
		je CSCL_NO		; Skip if there already

		; We must plant a "ADD DI,nnnn"

		push ax			; Save

		mov ax,0C781h		; "ADD DI,nn"
		call CSC_Poke_W		;

		pop ax			; Restore

		call CSC_Poke_W		; nnnn

		; Now, save this point for the undraw ?

CSCL_No:	mov cs:cOffset,dx	; Save the current position

		; Now we plant the undraw code here if required

		mov al,057h		; PUSH DI
		call CSC_Poke		;

		; Now, read the EGA/VGA to get the video data

;		mov ax,0258Ah		; MOV AH,[DI]
;		call CSC_Poke_W		;

;		mov cs:CSC_Flag_VR,1	; Flag a valid read

		; Now, we must plant the checking code here if required

;		nop			; Not yet

		; Now, we plant the successive masks

		mov bl,1		; The BL mask
		mov bh,01Dh		; "BL"

		call CSC_Plant_Byte	; Do it	

		mov bl,2		; The BH mask
		mov bh,03Dh		; "BH"

		call CSC_Plant_Byte	; Do it	

		mov bl,3		; The CL mask
		mov bh,00Dh		; "CL"

		call CSC_Plant_Byte	; Do it	

		mov bl,4		; The CH mask
		mov bh,02Dh		; "CH"

		call CSC_Plant_Byte	; Do it	

		; We have now driven this line. End.

		ret			; Done

; Given al = colour, ah = ext. byte, plant the mask and colour code

CSC_Plant_Byte:	push di			; Save

		mov ah,0		; Mask

		mov cx,8		; Count

		; Build the mask for this colour

CSCPB_Lp:	cmp bl,cs:[di]		; This pixel ?
		clc			; Assume no
		jne CSCPB_NTP		; No, skip

		stc			; Yes

CSCPB_NTP:	rcl ah,1		; Rotate it

		inc di			; Step ptr

		loop CSCPB_Lp		; Loop 8 times

		; ah = mask for this colour

		cmp ah,0		; Any ?
		je CSCPB_X		; No, skip

		; Now, we have a mask, do we already have this mask ?

		cmp b cs:CSC_Col_Mask,ah; Same mask ?
		je CSCPB_SkipMask	; Yes, skip

		; Yes, plant this mask and remember it for next time

		mov cs:CSC_Col_Mask,ah	; Save it for next time

		mov al,0B4h		; ax := "MOVE AH,mask"
		call CSC_Poke_W		; Poke it

		mov al,0EFh		; "OUT DX,AX"
		call CSC_Poke		;

		; Now, plant either "MOV [DI],nn" or "AND [DI],nn"

CSCPB_SkipMask:	mov al,088h		; "MOV"

		test b cs:CSC_Flag_VR,1	; Have we had a current read cycle ?
		jne CSCPB_VR		; Yes, skip		

		mov al,020h		; "AND"

CSCPB_VR:	call CSC_Poke		; Plant the instruction

		mov b cs:CSC_Flag_VR,0	; No valid read, OR next time

		; Now plant the EA byte

		mov al,bh		; EA byte/register select
		call CSC_Poke		;

		; We are done, exit

CSCPB_X:	pop di			; Restore ptr

		ret			; Done

; Data for generating successive instances of a sprite, shifted through the byte

Inst_Table:	dw Pix_Buff,	SC_0 - SC_Base	; Shifted no places
		dw Pix_Buff-1,	SC_1 - SC_Base	; 1
		dw Pix_Buff-2,	SC_2 - SC_Base	; 2
		dw Pix_Buff-3,	SC_3 - SC_Base	; 3
		dw Pix_Buff-4,	SC_4 - SC_Base	; 4
		dw Pix_Buff-5,	SC_5 - SC_Base	; 5
		dw Pix_Buff-6,	SC_6 - SC_Base	; 6
		dw Pix_Buff-7,	SC_7 - SC_Base	; 7

		dw 0,0			; End

; This is the header for the drawing code, copied onto the front of the routines

SC_Base:	cli			; DI

		push es			; Save old cs
		push ss			; Save ss

		pop es			;
		pop ss			;

		xchg sp,bp		; Use sp

		; Now, dx = clip value

		mov si,dx		; Save clip register

		mov dx,3CEh		; dx := port
		mov al,8		; al = BitMask

		sti			; EI again, using cs:Buffer as stack

		cmp ah,3		; 3 ? 
		jb SC_A			; No, 0,1,2

		je SC_3			; Is 3

		; 4,5,6,7

		cmp ah,5		; 5 ?
		jb SC_4			; Is 4

		je SC_5			; Is 5

		; 6,7

		cmp ah,6		; 6 ?
		je SC_6			; Is 6

		; 7

SC_7:		jmp SC_End		; Is 7

	; 0,1,2

SC_A:		cmp ah,1		; 1 ?
		ja SC_2			; Is 2

		je SC_1			; Is 1

SC_0:		jmp SC_End		; Is 0

		; Vectors

SC_1:		jmp SC_End		; Is 1

SC_2:		jmp SC_End		; Is 2

SC_3:		jmp SC_End		; Is 3

SC_4:		jmp SC_End		; Is 4

SC_5:		jmp SC_End		; Is 5

SC_6:		jmp SC_End		; Is 6

SC_End:		nop			;

; This is the header for each column

HC_Base:	inc si			; Step the clip count
		jns HC_End+2		; Skip if visible

		db 0E9h			; JMP (d16)

HC_End:		nop			; End

; This is the exit code

EX_Base:	cli			; DI

		push es			; Recover ss
		pop ss			;

		xchg sp,bp		; Restore sp/bp

		sti			; EI
		retf			; Exit

EX_End:		nop			; End

; Write ax to ds:si if flag set, but always inc si

CSC_Poke_W:	call CSC_Poke		; LSB

		xchg al,ah		;

		call CSC_Poke		; MSB

		xchg al,ah		;

		ret			; Done

; Write al to ds:si if flag set, but always inc si

CSC_Poke:	test b cs:CSC_Flag,1	; Write to memory ?
		jz CSCP_1		; No, skip

		mov [si],al		; Store it

CSCP_1:		inc si			; Skip

		ret			; Done

; Given di = ptr to a drawing, get it into the buffer and count the width and
; depth.

CSC_GetPixels:	call CSC_Clear		; Clear the pixel buffers

		; Now read it

		mov cx,0		; dX = 0 and dY = 0

		mov si,o Pix_Buff	; Point at the buffer

		; Get one line

CSC_GP_Y_Lp:	mov bx,si		; Save start of line

		; Read across the line

		mov ah,0		; None yet

CSC_GP_Lp:	mov al,cs:[di]		; Get the next item
		inc di			; Skip

		cmp al,0		; EOL ?
		je CSCGP_NL		; Yes, skip

		; No, save the pixel

		sub al,"0"		; Pixel ?
		jc CSC_GP_0		; No, leave as 0

		cmp al,3		; Pixel ?
		ja CSC_GP_0		; No, leave as 0

		; Valid, plant it

		inc al			; 1 .. 4

		mov cs:[si],al		; Store the pixel

CSC_GP_0:	inc si			; Skip it

		inc ah			; One more pixel

		; In range ?

		cmp ah,cscMax_X		; In range ?
		jae CSCGP_Fail		; No, fail

		; Ok, loop
		
		jmp s CSC_GP_Lp		; Loop

	; We are at the EOL

CSCGP_NL:	cmp cl,ah		; Is this higher ?
		jnc CSCGP_1		; No, skip

		mov cl,ah		; Save the new width

		; We have a line

CSCGP_1:	inc ch			; Step Y

		cmp ch,cscMax_Y		; In range ?
		jae CSCGP_Fail		; No, fail

		; Move ptr to the next line

		lea si,cscY_Step[bx]	; Move to start of the next line

		; Now, is there a next line ?

		mov al,cs:[di]		; End marker ?

		cmp al,0		; End ?
		jne CSC_GP_Y_Lp		; No, get another line

		; This is the end, skip marker

		inc di			; Skip

		mov cs:SC_Sizes,cx	; Save width and depth

		clc			; Ok		
		ret			; Done

	; General failure case

CSCGP_Fail:	stc			; Failed
		ret			;

; Clear the pixel buffers

CSC_Clear:	push di			; Save
		push es			;

		cld			; Clear the dirn. flag

		push cs			; es := cs
		pop es			;

		mov di,o Pix_Buff	; Clear it

		mov cx,(cscMax_X+8) * (cscMax_Y)	; Length
		mov al,0		; 0

	rep	stosb			; Clear the buffer

		pop es			; Restore
		pop di			;

		ret			; Done

; Vars

		align 2

DS_Rot		db ?			; Sprite rotate value

CSC_Flag	db 0			; Write to RAM flag

X_Off		db 0			; Offset from Top-Left to
Y_Off		db 0			;    the "sweet spot"

SC_Sizes	dw 0			; X,Y

cOffset		dw 0			; Offset to current
CSC_Flag_VR	db 0			; Flag if valid read cycle 

CSC_TOC		dw 0			; Ptr to column code start
CSC_Col_Code	db 0			; Code in this column

CSC_Col_Mask	db 0			; Current mask value

Large_Sprites	db 8 * nMax_Sprite dup (0)	; Sprite ptrs
Small_Sprites	db 8 * nMax_Sprite dup (0)	;

		db 8 dup (0)			; Buffer
Pix_Buff	db (cscY_Step) * (cscMax_Y) dup (0) ; Pixel buffer
